package edu.neu.ccs.assignment;

import edu.neu.ccs.bias.BiasCalculator;
import edu.neu.ccs.demeterf.TUCombiner;
import edu.neu.ccs.demeterf.demfgen.lib.List;
import edu.neu.ccs.util.ShannonCofactor;
import gen.Assignment;
import gen.Constraint;
import gen.Literal;
import gen.Neg;
import gen.Pos;
import gen.RawMaterial;
import gen.RawMaterialInstance;
import gen.RelationNr;
import gen.Variable;

/**
 * Class for generating best assignment and performing operations on it.
 * <p>
 * It uses Greedy approach for finding best assignment. 
 * For each variable present in the raw material:
 * <ol>
 * <li> Find E[Z[F(x=0)]]
 * <li> Find E[Z[F(x=1)]]
 * <li> Compare the above two values and pick max of the above two
 * <li> Set value of x in the raw material to the one that came out to be max by taking shannon cofactor
 * </ol>
 *  
 * @author Charu Chandra and Wu Jiang
 *
 */
public class AssignmentUtil {

	/**
	 * Finds the best assignment for the given raw material
	 * 
	 * @param rawMaterial
	 *            raw material in form of String
	 * @return best assignment as Assignment object
	 * @throws Exception
	 *             Any exception occurred during finding an assignment
	 */
	public static Assignment findAssignment(String rawMaterial)
	throws Exception {
		RawMaterial rm = RawMaterial.parse(rawMaterial);
		return findAssignment(rm);
	}

	/**
	 * Finds the best assignment for the given raw material
	 * 
	 * @param rawMaterial
	 *            raw material in form of RawMaterial
	 * @return best assignment as Assignment object
	 * @throws Exception
	 *             Any exception occurred during finding an assignment
	 */
	public static Assignment findAssignment(RawMaterial rawMaterial) throws Exception {

		List<Literal> result = null;
		// Get a list of all variables
		List<Variable> variables = listAllVars(rawMaterial);

		// comparing and finding the best assignment for each variable
		for (Variable v : variables) {
			double EZFx0 = sumExpectedValues(rawMaterial, v, 0);
			double EZFx1 = sumExpectedValues(rawMaterial, v, 1);

			Literal assignLit = (EZFx0 > EZFx1) ? new Literal(new Neg(), v)
			: new Literal(new Pos(), v);
			int value = (EZFx0 > EZFx1) ? 0 : 1;
			rawMaterial = updateRawMat(rawMaterial, v, value);
			if (result == null)
				result = List.create(assignLit);
			else
				result = result.append(assignLit);
		}
		return new Assignment(result);
	}

	/**
	 * List all the variables present in a given raw material
	 * 
	 * @param raw
	 * @return List of unique variables
	 */
	private static List<Variable> listAllVars(RawMaterial raw) {

		List<Variable> result = null;
		for (Constraint c : raw.instance.cs) {
			result = listAppend(result, c.vs);
		}
		return result;
	}

	/**
	 * Appends to list
	 * @param list
	 * @param vars
	 * @return
	 */
	private static List<Variable> listAppend(List<Variable> list,
			List<Variable> vars) {
		if (list == null)
			return vars;
		else {
			for (Variable var : vars) {
				if (!list.contains(var))
					list = list.append(var);
			}
		}

		return list;
	}

	/**
	 * Replaces all the relation numbers by their Shannon Cofactor 
	 * if variable v is set to value. 
	 * @param rm RawMaterial
	 * @param v Variable
	 * @param value 0 or 1
	 * @return Updated raw material
	 */
	private static RawMaterial updateRawMat(RawMaterial rm, Variable v,
			int value) {
		List<Constraint> list = rm.instance.cs;
		final int val = value;
		final Variable var = v;
		List<Constraint> result = list
		.map(new List.Map<Constraint, Constraint>() {
			public Constraint map(Constraint wc) {
				List<Variable> listVars = wc.vs;
				int posn = findPosition(listVars, var);
				if (posn == -1)
					return wc;
				return new Constraint(wc.w, new RelationNr(
						ShannonCofactor.computeSC(wc.r.v, posn, val)),
						listVars);
			}
		});

		return new RawMaterial(new RawMaterialInstance(result));
	}

	/**
	 * SUms the expected values for each of the row of RawMaterial 
	 * if variable v is set to value
	 * @param rm RawMaterial
	 * @param v Variable
	 * @param value 0 or 1
	 * @return Updated raw material
	 */
	private static double sumExpectedValues(RawMaterial rm, Variable v,
			int value) {
		final Variable var = v;
		final int val = value;

		Double sum = TUCombiner.traverse(rm, new TUCombiner<Double>() {

			public Double combine() {
				return 0.0d;
			}

			public Double fold(Double a, Double b) {
				return a.doubleValue() + b.doubleValue();
			}

			Double combine(Constraint wc) {
				int weight = wc.w.v;
				List<Variable> listVars = wc.vs;
				int posn = findPosition(listVars, var);
				if (posn == -1)
					return 0.0d;
				BiasCalculator biasCal = BiasCalculator.getInstance();

				double cal = biasCal.getPrice((ShannonCofactor
						.computeSC(wc.r.v, posn, val)));
				return weight * cal;
			}
		});

		return sum.doubleValue();
	}


	private static double sumExpectedValues(RawMaterial rm) {

		Double sum = TUCombiner.traverse(rm, new TUCombiner<Double>() {

			public Double combine() {
				return 0.0d;
			}

			public Double fold(Double a, Double b) {
				return a.doubleValue() + b.doubleValue();
			}

			Double combine(Constraint wc) {
				int weight = wc.w.v;
				double cal = BiasCalculator.getInstance().getPrice(wc.r.v);
				return weight * cal;
			}
		});


		return sum.doubleValue();
	}

	private static int findPosition(List<Variable> list, Variable var) {
		return list.index(var);
	}

	/**
	 * Calculates the quality of assignment for the input raw material.
	 * 
	 * @param rawMaterial
	 *            in form of String
	 * @param assign
	 *            in form of Assignment object
	 * @return quality achieved by the assignment
	 * @throws Exception
	 *             Any exception produced during calculation
	 */
	public static double findQuality(String rawMaterial, Assignment assign)
	throws Exception {
		RawMaterial rm = RawMaterial.parse(rawMaterial);
		return AssignmentVerifier.findQuality(rm, assign);
	}

	/**
	 * Calculates the quality of assignment for the input raw material.
	 * 
	 * @param rawMaterial
	 *            in form of RawMaterial object
	 * @param assign
	 *            in form of Assignment object
	 * @return quality achieved by the assignment
	 * @throws Exception
	 *             Any exception produced during calculation
	 */
	public static double findQuality(RawMaterial rawMaterial, Assignment assign)
	throws Exception {
		return AssignmentVerifier.findQuality(rawMaterial, assign);
	}
}
