package edu.neu.ccs.seller;

import edu.neu.ccs.assignment.AssignmentUtil;
import edu.neu.ccs.demeterf.demfgen.lib.List;
import edu.neu.ccs.rawmaterial.RawMaterialUtil;
import edu.neu.ccs.util.MinMaxSolver;
import gen.Assignment;
import gen.Derivative;
import gen.Player;
import gen.PlayerID;
import gen.Price;
import gen.RawMaterial;
import gen.RelationNr;
import gen.Type;
import gen.TypeInstance;

/**
 * Represents a Seller and provides all the operations that a seller might require for playing SDG. 
 * @author Charu Chandra and Wu Jiang
 *
 */
public class Seller extends Player{

	private PlayerID id = new PlayerID(6);

	//Constructor
	public Seller(int id, String name){
		super(new PlayerID(id), name);
	}

	/**
	 * Derivatives are generated randomly choosing two relations
	 * which gives profit of atleast 0.06
	 * @param dName
	 * @throws Exception
	 */
	public Derivative generateDerivative(String dName)throws Exception{
		Type type = getType(2);
		Price price = getPrice(type);
		String dervString = "deriv[ \"" + dName + "\" " + this.id.print() + " " + price.print() + " " + type.print()+ "]";
		return Derivative.parse(dervString);
	}

	/**
	 * Gets the optimum price for the given relations in form of Type. Currently it works for only two relations.
	 * @param type
	 * @return
	 */
	public Price getPrice(Type type){
		List<TypeInstance> instances = type.instances;
		double delta = 0.06d;
		return new Price(MinMaxSolver.getOptimumPrice(instances) + delta);
	}

	/**
	 * Gets the optimum price for the given derivative. Currently it works for only two relations.
	 * @param type
	 * @return
	 */
	public Price getPrice(Derivative d){
		return getPrice(d.type);
	}

	/**
	 * Gets Type containing relations for which the optimum price is less than 0.94 
	 * so that some profit could be made. 
	 * Currently it works for only two relations.
	 * @param noOfRel Number of relations to be produced. Currently produce only two relations.
	 * @return
	 */
	private Type getType(int noOfRel){
		RelationNr relation1 = null; 
		RelationNr relation2 = null;
		double price = 1.0d; 
		while(price >= 0.94d){
			int rel1 = ((int)(Math.random() * 1000) % 256);
			int rel2 = ((int)(Math.random() * 1000) % 256);
			relation1 = new RelationNr(rel1);
			relation2 = new RelationNr(rel2);
			price = MinMaxSolver.getOptimumPrice(relation1,relation2) + 0.05;
		}

		List<TypeInstance> instances = List.create(new TypeInstance(relation1));
		instances = instances.append(new TypeInstance(relation2));
		return new Type(instances);
	}

	/**
	 * Generates Raw material. It can currently generate raw material 
	 * for only for the first two relations.
	 * @param numVar - number of variables
	 * @param type - Type containing TypeInstance
	 * @return RawMaterial 
	 */
	public RawMaterial generateRawMaterial(int numVar, Type type) throws Exception{
		return RawMaterialUtil.generateRawMaterial(numVar, type);
	}

	/**
	 * Calculates the quality of assignment for the input raw material.
	 * 
	 * @param rawMaterial
	 *            in form of RawMaterial object
	 * @param assign
	 *            in form of Assignment object
	 * @return quality achieved by the assignment
	 * @throws Exception
	 *             Any exception produced during calculation
	 */
	public static double findQuality(RawMaterial rawMaterial, Assignment assign)
	throws Exception {
		return AssignmentUtil.findQuality(rawMaterial, assign);
	}
}
