package edu.neu.ccs.util;

import edu.neu.ccs.bias.BiasCalculator;
import edu.neu.ccs.demeterf.demfgen.lib.List;
import edu.neu.ccs.satsolver.PolynomialI;
import gen.RelationNr;
import gen.TypeInstance;

/**
 * Class for solving min max problem using numerical method
 * <p>
 * <ul>
 * <li> Objective function f = X * Pr[r1,b] + Y * Pr[r2,b], where X and Y are the weights for r1 and r2.
 * <li> Dividing by X+Y, we get f = X/X+Y * Pr[r1,b] + Y/X+Y * Pr[r2,b]
 * <li> Setting X/X+Y = a, f = a * Pr[r1,b] + (1-a) * Pr[r2,b]                 
 * </ul>
 * Since seller would want to minimize the satisfiability ratio and can set weights,
 * and buyer would want to maximize the satisfiability and can set b, 
 * the function f is minimized wrt a and maximized wrt b  
 * 
 * @author Charu Chandra and Wu Jiang
 *
 */
public class MinMaxSolver {

	/**
	 * Solves min max problem for two relations using numerical method.
	 * @param rel1
	 * @param rel2
	 * @return a 2 value array. 1st element is the optimum value of a = x/y and 
	 * 2nd is the optimum value of expected value
	 */
	public static double[] solveMinMax(int rel1, int rel2){
		// Find polynomial for relation1	
		PolynomialI prob1 = BiasCalculator.getInstance().getPolynomial(rel1);
		// Find polynomial for relation2
		PolynomialI prob2 = BiasCalculator.getInstance().getPolynomial(rel2);

		// f = a * prob1 + (1-a) * prob2
		double amin = -1;
		double bmax = -1;
		double fmin = Double.MAX_VALUE;

		for(double i = 0.0d; i <= 1.0d; i += 0.001d){
			// set values for a = x/(x+y)
			double a = i; 
			double fmax = Double.MIN_VALUE;
			double abmax = -1;
			for(double b = 0.0d; b <= 1.0d; b += 0.001d){
				double fn = a * prob1.eval(b) + (1-a) * prob2.eval(b);
				if(fn > fmax) {
					fmax = fn;
					abmax = b;
				}
			}

			double fn = a * prob1.eval(abmax) + (1-a) * prob2.eval(abmax);
			if(fn < fmin){
				fmin = fn;
				amin = a;
				bmax = abmax;
			}
		}
		double[] result = {amin,bmax,fmin};
		return result;
	}

	/**
	 * Calculate the optimum ratio x/x+y 
	 * where x - weight for relation1 and
	 *       y - weight for relation2 
	 * @param rel1
	 * @param rel2
	 * @return
	 */
	public static double getOptimumT(int rel1, int rel2){
		return solveMinMax(rel1, rel2)[0];
	}

	/**
	 * Calculate the optimum bias for the two relations 
	 * @param rel1
	 * @param rel2
	 * @return
	 */
	public static double getOptimumBias(int rel1, int rel2){
		return solveMinMax(rel1, rel2)[1];
	}
	/**
	 * Find the optimum price for input relations. Currently processes only two relations
	 * @param relations
	 * @return
	 */
	public static double getOptimumPrice(List<TypeInstance> relations){
		if(relations.length() == 0) return 1.0d;
		
		if(relations.length() == 1) 
			return BiasCalculator.getInstance().getPrice(relations.top().r.v);
		
		RelationNr relation1 = relations.top().r;
		RelationNr relation2 = relations.pop().top().r;
		return solveMinMax(relation1.v,relation2.v)[2];
	}
	/**
	 * Find the optimum price for two relations 
	 * @param relation1
	 * @param relation2
	 * @return
	 */
	public static double getOptimumPrice(RelationNr relation1, RelationNr relation2){		
		return solveMinMax(relation1.v,relation2.v)[2];
	}
}
