package edu.neu.ccs.assignment;

import edu.neu.ccs.demeterf.Bc;
import edu.neu.ccs.demeterf.TUCombiner;
import edu.neu.ccs.demeterf.demfgen.lib.List;
import edu.neu.ccs.demeterf.demfgen.lib.Map;
import edu.neu.ccs.util.ShannonCofactor;
import gen.Assignment;
import gen.Constraint;
import gen.Literal;
import gen.Pos;
import gen.RawMaterial;
import gen.RelationNr;
import gen.Variable;

/**
 * Class for verifying a given assignment
 * 
 * <p>
 * For each constraint, set each variable to the value specified in the assignment.
 * Setting variable can be done by taking shannon cofactor. After setting All the variables,
 * the final relation will come out to be 0 or 255. 0 means the constraint is not satisfied,
 * 255 means the constraint is not satisfied.
 * 
 * @author Charu Chandra and Wu Jiang
 *
 */
class AssignmentVerifier extends Bc {
    
    private static Assignment m_assign = null;

    //Comparator class
    private static class RelationNrComp implements java.util.Comparator<RelationNr> {
	public int compare(RelationNr a, RelationNr b) {
	    return (new Integer(a.v)).compareTo(new Integer(b.v));
	}
    }
    
    /**
     * Counts the number of row satisfied by the assignment in the given raw material
     * @param raw
     * @param assign
     * @return
     */
    static Map<RelationNr, Integer> findSatAssignments(RawMaterial raw,
	    Assignment assign) {
	m_assign = assign;
	Map<RelationNr, Integer> satMap = TUCombiner.traverse(raw,
		new TUCombiner<Map<RelationNr, Integer>>() {
		    Map<RelationNr, Integer> empty = Map
			    .<RelationNr, Integer> create(new RelationNrComp());

		    public Map<RelationNr, Integer> combine() {
			return empty;
		    }

		    public Map<RelationNr, Integer> fold(
			    Map<RelationNr, Integer> a,
			    Map<RelationNr, Integer> b) {
			return a.merge(b, new Map.Merge<Integer>() {
			    public Integer merge(Integer i, Integer j) {
				return i + j;
			    }
			});
		    }

		    Map<RelationNr, Integer> combine(Constraint wc) {
			return empty.put(computeReducedFunction(wc, m_assign),
				wc.w.v);
		    }
		});
	return satMap;
    }
    /**
     * Sets the variables as in the assignment on the constraint to check 
     * it is satisfied or not. If the resultant number is 255 if the constraint is 
     * satisfied or 0 if it is not satisfied. 
     * @param wc Constraint
     * @param assign Assignment
     * @return Returns the RelationNr containing the resultant relation number. 
     */
    private static RelationNr computeReducedFunction(Constraint wc, Assignment assign) {
	List<Literal> literals = assign.literals;

	int reducedRel = wc.r.v;
	int count = 0;
	for (Variable var : wc.vs) {
	    int value = 0;
	    for(Literal lit: literals){
		if(lit.var.equals(var)) {
		    value = (lit.value instanceof Pos)? 1:0; 
		}
	    }
	    reducedRel = ShannonCofactor.computeSC(reducedRel, count, value);
	    count ++;
	}
	
	return new RelationNr(reducedRel);
    }

    /**
     * Finds the quality of the assignment for the given raw material.
     * @param rm RawMaterial
     * @param assign Assignment
     * @return
     */
    static double findQuality(RawMaterial rm, Assignment assign) {
	Map<RelationNr, Integer> satMap = AssignmentVerifier
		.findSatAssignments(rm, assign);
	double satisfied = (satMap.containsKey(new RelationNr(255))) ? satMap
		.get(new RelationNr(255)) : 0;
	double unsatisfied = (satMap.containsKey(new RelationNr(0))) ? satMap
		.get(new RelationNr(0)) : 0;
	double total = satisfied + unsatisfied;
	return satisfied / total;
    }
}