package edu.neu.ccs.rawmaterial;
import edu.neu.ccs.evergreen.ir.Relation;
import edu.neu.ccs.util.Binomial;

/**
 * Class for generation of symmetric formula. It takes care of all types of relations:
 * <ol>
 * <li> symmetric relation - Relations for which arrangement of variables is not important.
 *  e.g relation 22 (001 or 010 or 100), a b c and c b a both can be satisfied by !a !b c. All the combinations 
 *  should be generated for these relations. Therefore, (n choose 3) constraint will be generated.
 * <li> asymmetric relation - Relations for which arrangement of variables is important.
 * e.g relation 2(001), a b c and c b a both can not be satisfied by same assignment. 1st requires
 * a !b !c and 2nd will be satisfied by c !b !a. In this case (n choose 3) * 6 constraints will be generated.
 * <li> dont cares - There are some relations that have terms that we dont care about. 
 * e.g. relation 170 (001 or 011 or 101 or 111) where only the last one matters, the other two 
 * are dont cares. In this case we introduce variable named dc1, dc2 to represent dont care terms.
 * The number of relations will depend on if the remaining formula is symmetric or asmmetric.    
 * </ol>
 * @author Charu Chandra and Wu Jiang
 */
class SymFormula {
	static int rankNo = 3; // rank for CSP formula
	static String  dc1 = "dc1";
	static String dc2 = "dc2";
	static String dc3 = "dc3";

	public static void p(String s) {
		System.out.println(s);
	}

	/**
	 * To check if the relation is symetric or mot
	 * @param relNum
	 * @return
	 */
	private static boolean isSymmetric(int relNum){
		// for rank =3;
		// Relation(int rank, int relationNo)
		Relation rel = new Relation(rankNo, relNum);
		int relevantNo = rel.numberOfRelevantVariables();
		int swapN01 = rel.swap(0, 1);
		int swapN12 = rel.swap(1, 2);
		int swapN02 = rel.swap(0, 2);

		if (relevantNo == 3){
			// swap all the three variables position, and compare the new relation number
			// Defn: A CSP formula is called symmetric if any permutation of the variables in the formula returns the same
			// formula up to a permutation of the constraints. [Ahmed Abdelmeged 2007, "The Evergreen Game: The Promise of
			// Polynomials to Boost Boolean MAX-CSP Solvers"]

			return (swapN01 == swapN12) && (swapN01 == swapN02) && (swapN01 == relNum);
		}
		else if (relevantNo == 2){
			if(rel.isIrrelevant(0) == true)
				return swapN12 == relNum;
			else if(rel.isIrrelevant(1) == true)
				return swapN02 == relNum;
			else return swapN01 ==relNum;

		}
		else return true;

	}

	private static boolean hasDontCares(int relNum){
		//for rank =3;
		// Relation(int rank, int relationNo)
		Relation rel = new Relation(rankNo, relNum);
		return !(rel.getRank() == rel.numberOfRelevantVariables());
	}

	// n variables, for symmetric relation number
	private static String[][] combinations(int n, int relNum){
		//for rank =3;
		// Relation(int rank, int relationNo)
		Relation rel = new Relation(rankNo, relNum);
		int relevantNo = rel.numberOfRelevantVariables();
		String[] var = new String[n];
		for(int i=0; i<var.length; i++){
			StringBuffer sb= new StringBuffer();
			var[i] = sb.append("X").append(i).toString();
			//p(var[i]);
		}

		int nCk = (int)Binomial.binomial(n, relevantNo);
		//p("nCk" + nCk);
		String[][] raw = new String[nCk][3];
		int step = 0;

		// relevantNum = 3;
		if(relevantNo==3){
			//for(int step = 0; step<nCk; step++){
			for(int first=0; first<(var.length-2); first++){
				for(int second=(first+1); second<(var.length-1); second++){
					for(int third=(second+1); third<var.length; third++){
						//p( var[first] +" " + var[second]+" " + var[third] + "\n");
						//p(var[first] +" " + var[second]+" " + var[third] + "\n");
						raw[step][0] = var[first];
						raw[step][1] = var[second];
						raw[step][2] = var[third];
						step = step+1;
					}
				}
			}//end for
		}


		// relevantNum = 2;
		else if(relevantNo ==2){
			if(rel.isIrrelevant(0) == true){
				for(int first=0; first<(var.length-1); first++){
					for(int second=(first+1); second<(var.length); second++){
						raw[step][0] = dc1;
						raw[step][1] = var[first];
						raw[step][2] = var[second];
						step = step+1;
					}
				}
			}
			else if(rel.isIrrelevant(1) == true){
				for(int first=0; first<(var.length-1); first++){
					for(int second=(first+1); second<(var.length); second++){
						raw[step][0] = var[first];
						raw[step][1] = dc2;
						raw[step][2] = var[second];
						step = step+1;
					}
				}
			}
			else 
				for(int first=0; first<(var.length-1); first++){
					for(int second=(first+1); second<(var.length); second++){
						raw[step][0] = var[first];
						raw[step][1] = var[second];
						raw[step][2] = dc3;
						step = step+1;
					}
				}
		}
		else if(relevantNo==1){
			if(rel.isIrrelevant(0)==rel.isIrrelevant(1)){
				for(int first=0; first<(var.length); first++){
					raw[step][0] = dc1;
					raw[step][1] = dc2;
					raw[step][2] = var[first];
					step = step+1;
				}
			}

			else if(rel.isIrrelevant(0)==rel.isIrrelevant(2)){
				for(int first=0; first<(var.length); first++){
					raw[step][0] = dc1;
					raw[step][1] = var[first];
					raw[step][2] = dc3;
					step = step+1;
				}
			}
			else 
				for(int first=0; first<(var.length); first++){
					raw[step][0] = var[first];
					raw[step][1] = dc2;
					raw[step][2] = dc3;
					step = step+1;
				}
		}

		else {
			raw[step][0] = dc1;
			raw[step][1] = dc2;
			raw[step][2] = dc3;
			step = step+1;
		}

		return raw;

	}

	/**
	 * Find all the permutations and combinations for a relation number 
	 * for n variables
	 * @param n
	 * @param relNum
	 * @return
	 */
	public static String[][] permAndCombinations(int n, int relNum){
		//		for rank =3;
		// Relation(int rank, int relationNo)
		Relation rel = new Relation(rankNo, relNum);
		int relevantNo = rel.numberOfRelevantVariables();
		String[] var = new String[n];
		for(int i=0; i<var.length; i++){
			StringBuffer sb= new StringBuffer();
			var[i] = sb.append("X").append(i).toString();
			//p(var[i]);
		}
		int nAk = (int)Binomial.nint(Math.exp(Binomial.logFactorial(relevantNo))); // relevantNo!
		int nCk = (int)Binomial.binomial(n, relevantNo);
		int nPk = nCk * nAk;
		//p("nCk " + nCk);
		String[][] raw = new String[nPk][3];


		String[][] raw_nCk = combinations(n, relNum);


		//relevantNum = 3;
		if(relevantNo==3){
			for(int i = 0; i< nCk; i++){
				raw[i][0] = raw_nCk[i][0];
				raw[i][1] = raw_nCk[i][1];
				raw[i][2] = raw_nCk[i][2];
			}
			for(int i = nCk; i < 2*nCk; i++){
				raw[i][0] = raw_nCk[i-nCk][0];
				raw[i][1] = raw_nCk[i-nCk][2];
				raw[i][2] = raw_nCk[i-nCk][1];
			}
			for(int i = 2*nCk; i < 3*nCk; i++){
				raw[i][0] = raw_nCk[i-2*nCk][1];
				raw[i][1] = raw_nCk[i-2*nCk][0];
				raw[i][2] = raw_nCk[i-2*nCk][2];
			}
			for(int i = 3*nCk; i < 4*nCk; i++){
				raw[i][0] = raw_nCk[i-3*nCk][1];
				raw[i][1] = raw_nCk[i-3*nCk][2];
				raw[i][2] = raw_nCk[i-3*nCk][0];
			}
			for(int i = 4*nCk; i < 5*nCk; i++){
				raw[i][0] = raw_nCk[i-4*nCk][2];
				raw[i][1] = raw_nCk[i-4*nCk][0];
				raw[i][2] = raw_nCk[i-4*nCk][1];
			}
			for(int i = 5*nCk; i < 6*nCk; i++){
				raw[i][0] = raw_nCk[i-5*nCk][2];
				raw[i][1] = raw_nCk[i-5*nCk][1];
				raw[i][2] = raw_nCk[i-5*nCk][0];
			}

		}

		// relevantNum = 2;
		if(relevantNo==2){
			if(rel.isIrrelevant(0)==true){
				for(int i = 0; i< nCk; i++){
					raw[i][0] = raw_nCk[i][0];
					raw[i][1] = raw_nCk[i][1];
					raw[i][2] = raw_nCk[i][2];
				}
				for(int i = nCk; i< 2*nCk; i++){
					raw[i][0] = raw_nCk[i-nCk][0];
					raw[i][1] = raw_nCk[i-nCk][2];
					raw[i][2] = raw_nCk[i-nCk][1];
				}
			}
			else if(rel.isIrrelevant(1)==true){
				for(int i = 0; i< nCk; i++){
					raw[i][0] = raw_nCk[i][0];
					raw[i][1] = raw_nCk[i][1];
					raw[i][2] = raw_nCk[i][2];
				}
				for(int i = nCk; i< 2*nCk; i++){
					raw[i][0] = raw_nCk[i-nCk][2];
					raw[i][1] = raw_nCk[i-nCk][1];
					raw[i][2] = raw_nCk[i-nCk][1];
				}
			}
			else {
				for(int i = 0; i< nCk; i++){
					raw[i][0] = raw_nCk[i][0];
					raw[i][1] = raw_nCk[i][1];
					raw[i][2] = raw_nCk[i][2];
				}
				for(int i = nCk; i< 2*nCk; i++){
					raw[i][0] = raw_nCk[i-nCk][1];
					raw[i][1] = raw_nCk[i-nCk][0];
					raw[i][2] = raw_nCk[i-nCk][2];
				}
			}

		}
		return raw;
	}

	/**
	 * Resturns a two dimensional array of Strings that contains rows of symmetric formula
	 * @param varNum
	 * @param relNum
	 * @return
	 */
	public static String[][] generateSymFormula(int varNum, int relNum){
		if (isSymmetric(relNum) == true){
			return combinations(varNum, relNum);
		}
		else return permAndCombinations(varNum, relNum);


	}	
}
